@@ -15,42 +15,8 @@ from pyRing.likelihood import project
 from pyRing.utils      import check_NR_dir, qnm_interpolate, qnm_interpolate_KN, resize_time_series, review_warning, set_prefix
 from pyRing            import waveform as wf
 
-def inject_ringdown_signal(lenstrain, tstart, length, ifo, triggertime, **kwargs):
 
-    """
-        Main function to set an injection using one of the analytical ringdown templates available.
-        Handles parameters common to all templates.
-    """
-    
-    time_wf = np.linspace(0,length,lenstrain+1)[:-1]
-
-    if (kwargs['injection-approximant']=='Damped-sinusoids'):
-        wf_model = damped_sinusoids_injection(**kwargs)
-    elif (kwargs['injection-approximant']=='Morlet-Gabor-wavelets'):
-        wf_model = morlet_gabor_wavelets_injection(**kwargs)
-    elif (kwargs['injection-approximant']=='Kerr'):
-        wf_model = kerr_injection(**kwargs)
-    elif (kwargs['injection-approximant']=='MMRDNS'):
-        wf_model = mmrdns_injection(**kwargs)
-    elif (kwargs['injection-approximant']=='MMRDNP'):
-        wf_model = mmrdnp_injection(**kwargs)
-    elif (kwargs['injection-approximant']=='TEOBResumSPM'):
-        wf_model = TEOBPM_injection(**kwargs)
-    elif (kwargs['injection-approximant']=='KHS_2012'):
-        wf_model = khs_injection(**kwargs)
-    hs, hvx, hvy, hp, hc = wf_model.waveform(time_wf)[0], wf_model.waveform(time_wf)[1], wf_model.waveform(time_wf)[2], wf_model.waveform(time_wf)[3], wf_model.waveform(time_wf)[4]
-
-    sys.stdout.write('* Injecting the `{}` waveform model in the {} detector.\n\n'.format(kwargs['injection-approximant'], ifo))
-
-    h        = np.zeros(len(hs))
-    srate    = kwargs['sampling-rate']
-    detector = lal.cached_detector_by_prefix[ifo]
-    ref_det  = lal.cached_detector_by_prefix[kwargs['ref-det']]
-    psi      = kwargs['injection-parameters']['psi']
-    tM_gps   = lal.LIGOTimeGPS(float(triggertime))
-    scaling  = kwargs['injection-scaling']
-    if not(scaling==1.0):
-        sys.stdout.write('* Applying a scaling factor {} to the injection.\n\n'.format(scaling))
+def load_injected_ra_dec(triggertime, kwargs):
 
     if (kwargs['sky-frame']=='detector'):
         tg, ra, dec = DetFrameToEquatorial(lal.cached_detector_by_prefix[kwargs['ref-det']],
@@ -64,15 +30,45 @@ def inject_ringdown_signal(lenstrain, tstart, length, ifo, triggertime, **kwargs
     else:
         raise ValueError("Invalid option for sky position sampling.")
 
-    hs, hvx, hvy, hp, hc = hs*scaling, hvx*scaling, hvy*scaling, hp*scaling, hc*scaling
-    wave       = project(hs, hvx, hvy, hp, hc, detector, ra, dec, psi, tM_gps)
+    return ra, dec
+
+def inject_ringdown_signal(times, triggertime, ifo, **kwargs):
+
+    """
+        Main function to set an injection using one of the analytical ringdown templates available.
+        Handles parameters common to all templates.
+    """
+    
+    # Compute the time-delay between detectors.
+    detector   = lal.cached_detector_by_prefix[ifo]
+    ref_det    = lal.cached_detector_by_prefix[kwargs['ref-det']]
+    psi        = kwargs['injection-parameters']['psi']
+    ra, dec    = load_injected_ra_dec(triggertime, kwargs)
+    tM_gps     = lal.LIGOTimeGPS(float(triggertime))
     time_delay = lal.ArrivalTimeDiff(detector.location, ref_det.location, ra, dec, tM_gps)
 
-    trig_idx   = math.ceil((triggertime+time_delay-tstart)*srate)
+    # Construct the waveform time axis. The zero (peak) of the signal is at `triggertime` by construction. Also, `triggertime` is equal to the time-axis middle value by construction.
+    time_axis_waveform = times - (triggertime+time_delay)
 
-    for i in range(0, len(wave)-trig_idx): h[trig_idx+i] = wave[i]
+    # Load signal injection.
+    if   (kwargs['injection-approximant']=='Damped-sinusoids'     ): wf_model = damped_sinusoids_injection(     **kwargs)
+    elif (kwargs['injection-approximant']=='Morlet-Gabor-wavelets'): wf_model = morlet_gabor_wavelets_injection(**kwargs)
+    elif (kwargs['injection-approximant']=='Kerr'                 ): wf_model = kerr_injection(                 **kwargs)
+    elif (kwargs['injection-approximant']=='MMRDNS'               ): wf_model = mmrdns_injection(               **kwargs)
+    elif (kwargs['injection-approximant']=='MMRDNP'               ): wf_model = mmrdnp_injection(               **kwargs)
+    elif (kwargs['injection-approximant']=='TEOBResumSPM'         ): wf_model = TEOBPM_injection(               **kwargs)
+    elif (kwargs['injection-approximant']=='KHS_2012'             ): wf_model = khs_injection(                  **kwargs)
+    
+    hs, hvx, hvy, hp, hc = wf_model.waveform(time_axis_waveform)[0], wf_model.waveform(time_axis_waveform)[1], wf_model.waveform(time_axis_waveform)[2], wf_model.waveform(time_axis_waveform)[3], wf_model.waveform(time_axis_waveform)[4]
+    sys.stdout.write('* Injecting the `{}` waveform model in the {} detector.\n\n'.format(kwargs['injection-approximant'], ifo))
 
-    return h
+    # Apply requested scaling and project waveform.
+    scaling = kwargs['injection-scaling']
+    if not(scaling==1.0): sys.stdout.write('* Applying a scaling factor of {} to the injection.\n\n'.format(scaling))
+    hs, hvx, hvy, hp, hc = hs*scaling, hvx*scaling, hvy*scaling, hp*scaling, hc*scaling
+    wave                 = project(hs, hvx, hvy, hp, hc, detector, ra, dec, psi, tM_gps)
+
+    return wave
 
 def damped_sinusoids_injection(**kwargs):
 
@@ -295,15 +291,17 @@ def khs_injection(**kwargs):
 
     return wf_model
 
-def inject_IMR_signal(lenstrain, tstart, length, ifo, triggertime, **kwargs):
+def inject_IMR_signal(times, triggertime, ifo, **kwargs):
 
     review_warning()
 
-    params  = lal.CreateDict()
-    deltaT  = 1.0/kwargs['sampling-rate']
-    f_ref   = kwargs['injection-parameters']['f-ref']
-    f_start = kwargs['injection-parameters']['f-start']
-    scaling = kwargs['injection-scaling']
+    lenstrain = len(times)
+    tstart    = times[0]
+    deltaT    = 1.0/kwargs['sampling-rate']
+    params    = lal.CreateDict()
+    f_ref     = kwargs['injection-parameters']['f-ref']
+    f_start   = kwargs['injection-parameters']['f-start']
+    scaling   = kwargs['injection-scaling']
 
     if(kwargs['injection-approximant']=='NR'):
         #=======================================================================================================================#
@@ -322,10 +320,6 @@ def inject_IMR_signal(lenstrain, tstart, length, ifo, triggertime, **kwargs):
         lalsim.SimInspiralWaveformParamsInsertPNAmplitudeOrder(params,int(kwargs['injection-parameters']['amp-order']))
         lalsim.SimInspiralWaveformParamsInsertPNPhaseOrder(params,int(kwargs['injection-parameters']['phase-order']))
 
-    detector  = lal.cached_detector_by_prefix[ifo]
-    ref_det   = lal.cached_detector_by_prefix[kwargs['ref-det']]
-    tM_gps    = lal.LIGOTimeGPS(float(triggertime))
-
     m1       = kwargs['injection-parameters']['m1']
     m2       = kwargs['injection-parameters']['m2']
     s1x      = kwargs['injection-parameters']['s1x_LALSim']
@@ -343,17 +337,10 @@ def inject_IMR_signal(lenstrain, tstart, length, ifo, triggertime, **kwargs):
         dist = dist/scaling
         sys.stdout.write('\n* Applying a scaling factor {} to the injection.'.format(scaling))
 
-    if (kwargs['sky-frame']=='detector'):
-        tg, ra, dec = DetFrameToEquatorial(lal.cached_detector_by_prefix[kwargs['ref-det']],
-                                           lal.cached_detector_by_prefix[kwargs['nonref-det']],
-                                           triggertime,
-                                           np.arccos(kwargs['injection-parameters']['cos_altitude']),
-                                           kwargs['injection-parameters']['azimuth'])
-    elif (kwargs['sky-frame']=='equatorial'):
-        ra  = kwargs['injection-parameters']['ra']
-        dec = kwargs['injection-parameters']['dec']
-    else:
-        raise ValueError("Invalid option for sky position sampling.")
+    detector   = lal.cached_detector_by_prefix[ifo]
+    ref_det    = lal.cached_detector_by_prefix[kwargs['ref-det']]
+    tM_gps     = lal.LIGOTimeGPS(float(triggertime))
+    ra, dec    = load_injected_ra_dec(triggertime, kwargs)
     time_delay = lal.ArrivalTimeDiff(detector.location, ref_det.location, ra, dec, tM_gps)
 
     if (kwargs['injection-parameters']['inject-modes'] is not None):
@@ -363,6 +350,7 @@ def inject_IMR_signal(lenstrain, tstart, length, ifo, triggertime, **kwargs):
             sys.stdout.write('l={}, m={}; \n'.format(mode[0], mode[1]))
             lalsim.SimInspiralModeArrayActivateMode(ModeArray, mode[0], mode[1])
         lalsim.SimInspiralWaveformParamsInsertModeArray(params, ModeArray)
+        sys.stdout.write('\n')
     elif (kwargs['injection-parameters']['inject-l-modes'] is not None):
         ModeArray = lalsim.SimInspiralCreateModeArray()
         sys.stdout.write('\n* Injecting a subset of l modes (all |m|<l modes are being injected): ')
@@ -370,6 +358,7 @@ def inject_IMR_signal(lenstrain, tstart, length, ifo, triggertime, **kwargs):
             sys.stdout.write('l={}; '.format(mode))
             lalsim.SimInspiralModeArrayActivateAllModesAtL(ModeArray, mode)
         lalsim.SimInspiralWaveformParamsInsertModeArray(params, ModeArray)
+        sys.stdout.write('\n')
     sys.stdout.write('\n')
 
     h_p, h_c = lalsim.SimInspiralChooseTDWaveform(m1*lal.MSUN_SI,
@@ -393,6 +382,6 @@ def inject_IMR_signal(lenstrain, tstart, length, ifo, triggertime, **kwargs):
 
     # Project the waveform onto a given detector.
     hs, hvx, hvy = np.zeros(len(hp)), np.zeros(len(hp)), np.zeros(len(hp))
-    h = project(hs, hvx, hvy, hp, hc, detector, ra, dec, psi, tM_gps)
+    h            = project(hs, hvx, hvy, hp, hc, detector, ra, dec, psi, tM_gps)
 
     return h
