@@ -29,21 +29,18 @@ except(ImportError, ModuleNotFoundError): pass
 # Data manipulation functions. #
 ################################
 
-def add_injection(on_source_strain, strain, starttime, T, ifo, triggertime, index_trigtime, signal_seglen, kwargs):
+def add_injection(on_source_times, on_source_strain, triggertime, ifo, kwargs):
 
-    # Inject the NR signal
-    if((kwargs['injection-approximant']=='NR') or ('LAL' in kwargs['injection-approximant'])): strain_injection = inject_IMR_signal(     strain.shape[0], starttime, T, ifo, triggertime = triggertime, **kwargs)
-    else                                                                                     : strain_injection = inject_ringdown_signal(strain.shape[0], starttime, T, ifo, triggertime = triggertime, **kwargs)
+    # Inject the signal
+    if((kwargs['injection-approximant']=='NR') or ('LAL' in kwargs['injection-approximant'])): on_source_injection = inject_IMR_signal(     on_source_times, triggertime, ifo, **kwargs)
+    else                                                                                     : on_source_injection = inject_ringdown_signal(on_source_times, triggertime, ifo, **kwargs)
 
-    # Pass as output for the strain the on-source chunk.
-    if not((signal_seglen%2)==0): on_source_injection = strain_injection[index_trigtime-signal_seglen//2:index_trigtime+signal_seglen//2+1]
-    else:                         on_source_injection = strain_injection[index_trigtime-signal_seglen//2:index_trigtime+signal_seglen//2]
     # Add noise on top of injection
     if not(kwargs['zero-noise']):
-        sys.stdout.write('* Adding noise on top of injection.\n\n')
+        sys.stdout.write('\n* Adding noise on top of injection.\n\n')
         on_source_strain += on_source_injection
     else:
-        sys.stdout.write('* Zero noise injection selected.\n\n')
+        sys.stdout.write('\n* Zero noise injection selected.\n\n')
         on_source_strain  = on_source_injection
 
     return on_source_strain
@@ -114,11 +111,14 @@ def bandpass_data(rawstrain, f_min_bp, f_max_bp, srate_dt, bandpassing_flag):
     return strain
 
 def check_sampling_rate_compatibility(requested_sampling_rate, data_sampling_rate):
+    
     if (requested_sampling_rate > data_sampling_rate): raise ValueError("* You requested a sampling rate higher than the data sampling.")
     else                                             : return
 
 def check_seglen_compatibility(signal_chunk_size, noise_chunk_size, T):
+    
     assert not((noise_chunk_size > T) or (signal_chunk_size > T)), "* Noise ({} s) and signal ({} s) seglens must be shorter than data duration ({})".format(noise_chunk_size, signal_chunk_size, T)
+    
     return
 
 def downsample_data(strain, srate, srate_dt):
@@ -157,19 +157,21 @@ def chunks(times, strain, chunksize, trigger_time):
 
     return np.array(time_chunks_list), np.array(data_chunks_list), index_trig
 
-def compute_on_off_source_strain(strain, signal_seglen, index_trigtime):
+def compute_on_off_source_strain(times, strain, signal_seglen, index_trigtime):
 
     on_source_mask = np.ones(len(strain), dtype=bool)
     if not((signal_seglen%2)==0):
         on_source_strain = strain[index_trigtime-signal_seglen//2:index_trigtime+signal_seglen//2+1]
+        on_source_times = times[index_trigtime-signal_seglen//2:index_trigtime+signal_seglen//2+1]
         on_source_mask[range(index_trigtime-signal_seglen//2,index_trigtime+signal_seglen//2+1,1)] = False
     else:
         on_source_strain = strain[index_trigtime-signal_seglen//2:index_trigtime+signal_seglen//2]
+        on_source_times = times[index_trigtime-signal_seglen//2:index_trigtime+signal_seglen//2]
         on_source_mask[range(index_trigtime-signal_seglen//2,index_trigtime+signal_seglen//2,1)]   = False
 
     off_source_strain = tuple((strain)[on_source_mask])
 
-    return on_source_strain, off_source_strain
+    return on_source_times, on_source_strain, off_source_strain
 
 def window_onsource_strain(on_source_strain, signal_seglen, noise_seglen, window_onsource_flag, window_flag, alpha_window, truncate):
 
@@ -187,6 +189,18 @@ def check_chunksizes_consistency(signal_chunk_size, noise_chunk_size, truncate):
 
     return
 
+def set_random_seed(user_seed, ifo):
+    
+    # If requested by the user, fix the noise seed.
+    if not(user_seed==-1):
+        
+        if(  ifo=='H1'): np.random.seed(user_seed)
+        elif(ifo=='L1'): np.random.seed(user_seed+1)
+        elif(ifo=='V1'): np.random.seed(user_seed+2)
+        else           : raise ValueError("Noise generation for this detector not supported. Please add the detector to the noise generation.")
+            
+    return
+
 ##########################
 # ACF related functions. #
 ##########################
@@ -544,13 +558,57 @@ def check_data_gaussianity(times, data, Covariance, signal_seglen, trigger_time,
     plt.legend(loc='upper right')
     plt.savefig(os.path.join(outdir, 'Noise',' scatter_sigmas_whitened_data_{}.pdf'.format(string_output)), bbox_inches='tight')
 
-    print('* A Kolmogorov–Smirnov test of whitened data chunks gave {}/{} normality outliers (at {} % significance).'.format(N_outliers, len_tot, p_value_threshold*100))
+    print('* A Kolmogorov–Smirnov test of whitened data chunks gave {}/{} normality outliers (at {} % significance).\n'.format(N_outliers, len_tot, p_value_threshold*100))
 
     return
 
-###########################
-# Loading data functions. #
-###########################
+########################################
+# {Loading/Generation} data functions. #
+########################################
+
+def check_consistency_loading_options(ifo_datafile, download_data_flag, kwargs):
+
+    if((not(ifo_datafile=='') and download_data_flag) or (download_data_flag and not(kwargs['gaussian-noise']=='')) or (not(ifo_datafile=='') and not(kwargs['gaussian-noise']==''))): raise ValueError("Contrasting options were passed for data loading. The three options: datafile={}, download-data={} and gaussian-noise={} cannot have more than one simultaneous non-zero value.".format(ifo_datafile, download_data_flag, kwargs['gaussian-noise']))
+
+    return
+
+def download_data_with_gwdatafind(ifo, starttime, endtime, channel, kwargs):
+    
+    if(kwargs['gw-data-type-{}'.format(ifo)]==''): raise ValueError('Data download method is gwdatafind, but no `gw-data-type-{}` option passed.'.format(ifo))
+    
+    connection = datafind.GWDataFindHTTPConnection()
+    cache      = connection.find_frame_urls(ifo[0], kwargs['gw-data-type-{}'.format(ifo)], starttime, endtime, urltype='file')
+    tseries    = TimeSeries.read(cache, channel)
+
+    return tseries
+
+def download_data_with_gwpy(ifo, starttime, endtime, channel, kwargs):
+
+    sys.stdout.write('\n* Using GWPY to download data.\n\n')
+    tag        = kwargs['tag']
+    tseries    = fetch_data(ifo, starttime, endtime, channel=channel, path=None, verbose=2, tag=tag)
+
+    return tseries
+
+def download_data(ifo, triggertime, kwargs):
+    
+    T         = float(kwargs['datalen-download'])
+    starttime = int(triggertime)-(T/2.)
+    endtime   = int(triggertime)+(T/2.)
+    channel   = kwargs['channel-{}'.format(ifo)]
+
+    if(kwargs['gw-data-find']==1): tseries = download_data_with_gwdatafind(ifo, starttime, endtime, channel, kwargs)
+    else                         : tseries = download_data_with_gwpy(      ifo, starttime, endtime, channel, kwargs)
+
+    # The `starttime` variable previously defined can be off by as much as one sample compared to the first sample of the data. Here, we re-set it to the actual
+    starttime = tseries.t0.to_value()
+    dt        = tseries.dt.to_value()
+    srate_dt  = tseries.sample_rate.to_value()
+    rawstrain = np.array(tseries.data)
+    
+    sys.stdout.write('\n* Loading {}s of data from channel {} starting at {}.\n\n'.format(T, channel, starttime))
+    
+    return starttime, dt, srate_dt, T, rawstrain
 
 def fetch_data(ifo, tstart, tend, channel=None, path=None, verbose=0, tag=None):
 
@@ -584,6 +642,158 @@ def fetch_data(ifo, tstart, tend, channel=None, path=None, verbose=0, tag=None):
 
     return tseries
 
+def generate_gaussian_noise(ifo, triggertime, srate, kwargs):
+    
+    set_random_seed(kwargs['gaussian-noise-seed'], ifo)
+
+    # Time axis definition.
+    T         = int(kwargs['injection-T'])
+    starttime = triggertime - T/2.
+    dt        = 1.0/srate
+    srate_dt  = srate
+
+    # Noise strain {generation/reading}.
+    if not(kwargs['run-type']=='post-processing'):
+        if  (kwargs['gaussian-noise']=='white'     ): rawstrain = generate_white_gaussian_noise(kwargs['gaussian-noise-white-sigma'], srate, T)
+        elif('coloured' in kwargs['gaussian-noise']): rawstrain = generate_coloured_gaussian_noise(ifo, srate, dt, T, kwargs)
+        else                                        : raise ValueError("* Unknown gaussian noise option selected.")
+        np.savetxt(os.path.join(kwargs['output'],'Noise','rawstrain_gaussian_noise_{}_{:d}_{:d}_{:d}.txt'.format(ifo, int(starttime), int(T), int(srate))), rawstrain)
+    else:
+        sys.stdout.write('* Reading the strain previously generated with gaussian noise.\n\n')
+        rawstrain = np.loadtxt(os.path.join(kwargs['output'],'Noise','rawstrain_gaussian_noise_{}_{:d}_{:d}_{:d}.txt'.format(ifo, int(starttime), int(T), int(srate))))
+
+    return starttime, dt, srate_dt, T, rawstrain
+
+def generate_white_gaussian_noise(sigma, srate, T):
+    
+    sys.stdout.write('* Generating white gaussian noise with zero mean and sigma = {}.\n\n'.format(sigma))
+    noise_strain = np.random.normal(loc=0.0, scale=sigma, size=int(T*srate))
+
+    return noise_strain
+
+def generate_coloured_gaussian_noise_from_acf_file(acf_file):
+    
+    times_cgn, ACF_cgn = np.loadtxt(acf_file, unpack=True)
+    C_cgn              = toeplitz(ACF_cgn)
+    cgn                = multivariate_normal(mean = np.zeros(C_cgn.shape[0]), cov = C_cgn)
+    noise_strain       = cgn.rvs()
+
+    return noise_strain
+
+def generate_coloured_gaussian_noise_from_psd_file(psd_file, srate, dt, T, kwargs):
+
+    # Read file and adapt to conventions.
+    sys.stdout.write('* Generating coloured gaussian noise with zero mean and PSD given by `{}`.\n\n'.format(psd_file))
+    freqs_file, psd_cgn = np.loadtxt(psd_file, unpack=True)
+    if('ASD' in psd_file):
+        sys.stdout.write("* Warning: a file containing the `ASD` word in the name was passed, thus the values contained in the file are being squared to compute the PSD.\n\n")
+        psd_cgn = psd_cgn*psd_cgn
+    
+    # Auxiliary quantities.
+    N_points  = int(T*srate)
+    freqs_cgn = np.fft.rfftfreq(N_points, d = dt)
+    df_cgn    = np.diff(freqs_cgn)[0]
+
+    # Interpolate PSD.
+    sys.stdout.write("* Interpolating the PSD passed in input on the frequency axis defined by the selected time segment, using `scipy.interpolate.interp1d`. The interpolation is using the `fill_value='extrapolate' and `bounds_error=False` methods to treat the boundaries.`\n\n")
+    psd_cgn = interp1d(freqs_file, psd_cgn, fill_value='extrapolate', bounds_error=False)
+
+    if(kwargs['gaussian-noise'] == 'coloured-TD'):
+        
+        # FIXME: when chosen fixing a gaussian-noise-seed, it gives a MemoryError. To be tested.
+        review_warning()
+        sys.stdout.write('* Generating the noise in time domain.\n\n')
+        
+        # We are using the one-sided PSD, thus it is twice the Fourier transform of the autocorrelation function (see eq. 7.15 of Maggiore Vol.1). We take the real part just to convert the complex output of fft to a real numpy float. The imaginary part if already 0 when coming out of the fft.
+        ACF_cgn            = 0.5*np.real(np.fft.irfft(psd_cgn(freqs_cgn)*df_cgn))*N_points
+        C_cgn              = toeplitz(ACF_cgn)
+        cgn                = multivariate_normal(mean = np.zeros(C_cgn.shape[0]), cov = C_cgn)
+        rawstrain          = cgn.rvs()
+
+    elif(kwargs['gaussian-noise'] == 'coloured-FD'):
+        
+        f_min_hardbound  = 11.0
+        f_max_hardbound  = 4096.
+        f_min_inj        = np.max([f_min_hardbound, freqs_cgn.min()])
+        f_max_inj        = np.min([f_max_hardbound, freqs_cgn.max()])
+        kmin             = int(f_min_inj/df_cgn)
+        kmax             = int(f_max_inj/df_cgn)
+        frequencies      = df_cgn*np.arange(0,N_points/2.+1)
+        frequency_strain = np.zeros(len(frequencies), dtype = np.complex64)
+        # Generate the frequency axis. The [f_min_hardbound, f_max_hardbound] Hz interval is chosen to be the range in which LIGO-Virgo-Kagra noise is under control. Exact values are unimportant, since the PSD will be estimated from the generated strain after bandpassing is potentially applied.
+        sys.stdout.write('* Generating the noise in frequency domain in the interval [{}, {}] Hz. This interval will later be shrinked according to bandpassing options.\n\n'.format(f_min_hardbound, f_max_hardbound))
+        
+        # Generate the noise in the frequency domain.
+        for i in range(kmin, kmax+1):
+            sigma_cgn           = 0.5*np.sqrt(psd_cgn(frequencies[i])/df_cgn)
+            frequency_strain[i] = np.random.normal(0.0,sigma_cgn) + 1j*np.random.normal(0.0,sigma_cgn)
+        
+        # Convert to time domain.
+        noise_strain = np.real(np.fft.irfft(frequency_strain))*df_cgn*N_points
+
+    else:
+        raise ValueError("* To generate gaussian noise, the allowed options are: 'white', 'coloured-TD', 'coloured-FD'.")
+
+    return noise_strain
+
+def generate_coloured_gaussian_noise(ifo, srate, dt, T, kwargs):
+
+    if   not(kwargs['acf-{}'.format(ifo)]==''): noise_strain = generate_coloured_gaussian_noise_from_acf_file(kwargs['acf-{}'.format(ifo)])
+    elif not(kwargs['psd-{}'.format(ifo)]==''): noise_strain = generate_coloured_gaussian_noise_from_psd_file(kwargs['psd-{}'.format(ifo)], srate, dt, T, kwargs)
+    else                                      : raise Exception("* If coloured gaussian noise is selected, an ACF or PSD from which the noise should be generated must be passed in input.")
+
+    return noise_strain
+
+def read_data_from_custom_file(fname):
+    
+    # If the file name does not follow GWOSC conventions, times corresponding to the given strain must also be passed in the file (this could be made more general by passing custom {starttime, srate, T}, but this way is probably more stable).
+    
+    times, rawstrain = np.loadtxt(fname, unpack=True)
+    starttime        = float(times[0])
+    dt               = float(times[1] - times[0])
+    T                = dt * len(rawstrain)
+
+    return starttime, dt, T, rawstrain
+
+def read_data_from_txt_file(fname):
+
+    ifo_name, fr_type, starttime, T = ((fname.split('/'))[-1]).strip('.txt').split('-')
+    rawstrain                       = np.loadtxt(fname)
+
+    return starttime, T, rawstrain
+
+def read_data_from_gwf_file(fname):
+    
+    ifo_name, fr_type, starttime, T = ((fname.split('/'))[-1]).strip('.gwf').split('-')
+    channel   = kwargs['channel-{}'.format(ifo)]
+    tseries   = TimeSeries.read(fname, channel)
+    rawstrain = np.array(tseries.data)
+
+    return starttime, T, rawstrain
+
+def read_data_from_GWOSC_file(fname, ifo, kwargs):
+
+    sys.stdout.write("* Warning: The file name is expected to follow GWOSC conventions 'DET-FRAMETYPE-STARTTIME--DATALEN.txt', e.g.: 'H-H1_GWOSC_4_V1-1126259446-32.txt'. See https://www.gw-openscience.org for more infomation. This requirement can be relaxed by activating the 'ignore-data-filename' option.\n")
+    assert not(fname==''), "Data are empty. Either pass valid data or select one of the 'download-data' or 'gaussian-noise' options."
+    
+    if not('.gwf' in fname): starttime, T, rawstrain = read_data_from_txt_file(fname)
+    else                   : starttime, T, rawstrain = read_data_from_gwf_file(fname, ifo, kwargs)
+    
+    starttime = float(starttime)
+    T         = float(T)
+    dt        = T/len(rawstrain)
+
+    return starttime, dt, T, rawstrain
+
+def read_data_from_file(fname, ifo, kwargs):
+        
+    if not(kwargs['ignore-data-filename']): starttime, dt, T, rawstrain = read_data_from_GWOSC_file(fname, ifo, kwargs)
+    else                                  : starttime, dt, T, rawstrain = read_data_from_custom_file(fname)
+    
+    srate_dt = 1./dt
+    sys.stdout.write('\n* Loaded `{}` starting at `{}` length {}s.\n\n'.format(fname,starttime,T))
+
+    return starttime, dt, srate_dt, T, rawstrain
 
 #FIXME: this function is too big and should be split. Also, it should be a class or a subclass of the detector class, not a function
 def load_data(ifo, fname, **kwargs):
@@ -637,122 +847,13 @@ def load_data(ifo, fname, **kwargs):
     # Data loading/generation section. #
     ####################################
 
-    # Start to read the data.
-    if not(ifo_datafile==''):
-        # If a data file is passed, check that no contrasting option (generate gaussian-noise or download data) is passed.
-        if (not(kwargs['gaussian-noise']=='')): raise ValueError("If 'gaussian-noise' is selected, no data must be passed.")
-        else:
-            if not(kwargs['ignore-data-filename']):
-                sys.stdout.write("* Warning: The file name is expected to follow GWOSC conventions 'DET-FRAMETYPE-STARTTIME--DATALEN.txt', e.g.: 'H-H1_GWOSC_4_V1-1126259446-32.txt'. See https://www.gw-openscience.org for more infomation. This requirement can be relaxed by activating the 'ignore-data-filename' option.\n")
-                assert not(fname==''), "Data are empty. Either pass valid data or select one of the 'download-data' or 'gaussian-noise' options."
-                if not('.gwf' in fname):
-                    ifo_name, fr_type, starttime, T = ((fname.split('/'))[-1]).strip('.txt').split('-')
-                    rawstrain = np.loadtxt(fname)
-                else:
-                    ifo_name, fr_type, starttime, T = ((fname.split('/'))[-1]).strip('.gwf').split('-')
-                    channel   = kwargs['channel-{}'.format(ifo)]
-                    tseries   = TimeSeries.read(fname, channel)
-                    rawstrain = np.array(tseries.data)
-                starttime     = float(starttime)
-                T             = float(T)
-                dt            = T/len(rawstrain)
-            else:
-                # If the file name does not follow GWOSC conventions, times corresponding to the given strain must also be passed in the file (this could be made more general by passing custom {starttime, srate, T}, but this way is probably more stable).
-                times, rawstrain = np.loadtxt(fname, unpack=True)
-                starttime        = float(times[0])
-                dt               = float(times[1] - times[0])
-                T                = dt * len(rawstrain)
-            srate_dt = 1./dt
-            sys.stdout.write('\n* Loaded `{}` starting at `{}` length {}s.\n\n'.format(fname,starttime,T))
-    else:
-        # If no data file was passed either download them or generate them yourself.
-        if(download_data_flag):
-            T         = float(kwargs['datalen-download'])
-            channel   = kwargs['channel-{}'.format(ifo)]
-            starttime = int(triggertime)-(T/2.)
-            endtime   = int(triggertime)+(T/2.)
+    # Consistency check.
+    check_consistency_loading_options(ifo_datafile, download_data_flag, kwargs)
 
-            if((kwargs['gw-data-find']==1) and not(kwargs['gw-data-type-{}'.format(ifo)]=='')):
-                connection = datafind.GWDataFindHTTPConnection()
-                cache      = connection.find_frame_urls(ifo[0], kwargs['gw-data-type-{}'.format(ifo)], starttime, endtime, urltype='file')
-                tseries    = TimeSeries.read(cache, channel)
-            else:
-                sys.stdout.write('\n* Using GWPY to download data.\n\n')
-                tag        = kwargs['tag']
-                tseries    = fetch_data(ifo, starttime, endtime, channel=channel, path=None, verbose=2, tag=tag)
-            rawstrain               = np.array(tseries.data)
-            dt                      = tseries.dt.to_value()
-            srate_dt                = tseries.sample_rate.to_value()
-            sys.stdout.write('\n* Loaded channel {} starting at {} length {}s.\n\n'.format(channel,starttime,T))
-        elif(not(kwargs['gaussian-noise']=='')):
-
-            # If requested by the user, fix the noise seed.
-            if not(kwargs['gaussian-noise-seed']==-1):
-                if(  ifo=='H1'): np.random.seed(kwargs['gaussian-noise-seed'])
-                elif(ifo=='L1'): np.random.seed(kwargs['gaussian-noise-seed']+1)
-                elif(ifo=='V1'): np.random.seed(kwargs['gaussian-noise-seed']+2)
-                else: raise ValueError("Noise generation for this detector not supported. Please add the detector to the noise generation.")
-
-            T            = int(kwargs['injection-T'])
-            starttime    = triggertime - T/2.
-            dt           = 1.0/srate
-            srate_dt     = srate
-            if not(kwargs['run-type']=='post-processing'):
-                if(kwargs['gaussian-noise']=='white'):
-                    sys.stdout.write('* Generating white gaussian noise with zero mean and sigma = {}.\n\n'.format(kwargs['gaussian-noise-white-sigma']))
-                    rawstrain = np.random.normal(loc=0.0, scale=kwargs['gaussian-noise-white-sigma'], size=int(T*srate))
-                elif('coloured' in kwargs['gaussian-noise']):
-                    assert (signal_seglen==noise_seglen), "The injected psd is being interpolated on a frequency set determined by the noise seglen. A consistent signal seglen is required to obtain an unbiased estimation."
-                    if not(kwargs['acf-{}'.format(ifo)]==''):
-                        times_cgn, ACF_cgn  = kwargs['acf-{}'.format(ifo)]
-                        C_cgn               = toeplitz(ACF_cgn)
-                        cgn                 = multivariate_normal(mean = np.zeros(C_cgn.shape[0]), cov = C_cgn)
-                        rawstrain           = cgn.rvs()
-                    elif not(kwargs['psd-{}'.format(ifo)]==''):
-                        psd_file            = kwargs['psd-{}'.format(ifo)]
-                        sys.stdout.write('* Generating coloured gaussian noise with zero mean and PSD given by `{}`.\n\n'.format(psd_file))
-                        freqs_file, psd_cgn = np.loadtxt(psd_file, unpack=True)
-                        if('ASD' in psd_file):
-                            print("* An ASD was passed, thus the ASD is being squared to compute the PSD.\n")
-                            psd_cgn = psd_cgn*psd_cgn
-                        N_points  = int(T*srate)
-                        psd_cgn   = interp1d(freqs_file, psd_cgn, fill_value='extrapolate', bounds_error=False)
-                        freqs_cgn = np.fft.rfftfreq(N_points, d = dt)
-                        df_cgn    = np.diff(freqs_cgn)[0]
-                        if(kwargs['gaussian-noise'] == 'coloured-TD'):
-                            # FIXME: when chosen fixing a gaussian-noise-seed, it gives a MemoryError. To be tested.
-                            review_warning()
-                            sys.stdout.write('* Generating the noise in time domain.\n\n')
-                            # We are using the one-sided PSD, thus it is twice the Fourier transform of the autocorrelation function (see eq. 7.15 of Maggiore Vol.1). We take the real part just to convert the complex output of fft to a real numpy float. The imaginary part if already 0 when coming out of the fft.
-                            ACF_cgn            = 0.5*np.real(np.fft.irfft(psd_cgn(freqs_cgn)*df_cgn))*N_points
-                            C_cgn              = toeplitz(ACF_cgn)
-                            cgn                = multivariate_normal(mean = np.zeros(C_cgn.shape[0]), cov = C_cgn)
-                            rawstrain          = cgn.rvs()
-                        elif(kwargs['gaussian-noise'] == 'coloured-FD'):
-                            sys.stdout.write('* Generating the noise in frequency domain.\n\n')
-                            # Range in which LIGO-Virgo noise is under control. Exact values are unimportant, since the PSD will be estimated from the generated strain after bandpassing is potentially applied.
-                            f_min_inj = np.max([11.0, freqs_cgn.min()])
-                            f_max_inj = np.min([4096., freqs_cgn.max()])
-                            kmin = int(f_min_inj/df_cgn)
-                            kmax = int(f_max_inj/df_cgn)
-                            frequencies      = df_cgn*np.arange(0,N_points/2.+1)
-                            frequency_strain = np.zeros(len(frequencies), dtype = np.complex64)
-                            for i in range(kmin, kmax+1):
-                                sigma_cgn = 0.5*np.sqrt(psd_cgn(frequencies[i])/df_cgn)
-                                frequency_strain[i] = np.random.normal(0.0,sigma_cgn)+1j*np.random.normal(0.0,sigma_cgn)
-                            rawstrain = np.real(np.fft.irfft(frequency_strain))*df_cgn*N_points
-                        else:
-                            raise ValueError("* To generate gaussian noise, the allowed options are: 'white', 'coloured-TD', 'coloured-FD'.")
-                    else:
-                        raise Exception("* If coloured gaussian noise is selected, an ACF or PSD from which the noise should be generated must be passed in input.")
-                else:
-                    raise ValueError("* Unknown gaussian noise option selected.")
-                np.savetxt(os.path.join(kwargs['output'],'Noise','rawstrain_gaussian_noise_{}_{:d}_{:d}_{:d}.txt'.format(ifo, int(starttime), int(T), int(srate))), rawstrain)
-            else:
-                sys.stdout.write('* Reading the strain previously generated with gaussian noise.\n\n')
-                rawstrain = np.loadtxt(os.path.join(kwargs['output'],'Noise','rawstrain_gaussian_noise_{}_{:d}_{:d}_{:d}.txt'.format(ifo, int(starttime), int(T), int(srate))))
-        else:
-            raise Exception("* Either pass a data file, download data or generate data.")
+    if not(ifo_datafile=='')               : starttime, dt, srate_dt, T, rawstrain = read_data_from_file(fname, ifo, kwargs)
+    elif(download_data_flag)               : starttime, dt, srate_dt, T, rawstrain = download_data(ifo, triggertime, kwargs)
+    elif(not(kwargs['gaussian-noise']=='')): starttime, dt, srate_dt, T, rawstrain = generate_gaussian_noise(ifo, triggertime, srate, kwargs)
+    else                                   : raise Exception("* Either pass a data file, download data or generate data.")
 
     # Safety checks.
     check_seglen_compatibility(signal_chunk_size, noise_chunk_size, T)
@@ -775,12 +876,10 @@ def load_data(ifo, fname, **kwargs):
     # Construct the time axis and compute the index of the trigtime (i.e. the estimate of the coalescence time of a signal or the time at which the injection should be placed). Then compute the closest datapoint to the requested trigtime, this is the true time sample corresponding to the trigtime on our discretised axis. This adds a maximum shift to the requested starttime of `0.5*dt`.
     times            = np.linspace(starttime, starttime+T-dt, len(strain))
     index_trigtime   = round((triggertime-starttime)*srate)
-    triggertime_data = times[index_trigtime]
-
-    # Find the on-source chunk, centered on the trigger time, and the off-source chunk to avoid including the signal in the PSD computation.
-    # Since the ACF might be computed on a windowed chunk through an fft, the onsource chunk must be windowed when no truncation is applied and the fft requested, to avoid normalisation inconsistencies.
-    on_source_strain, off_source_strain = compute_on_off_source_strain(strain, signal_seglen, index_trigtime)
-    on_source_strain                    = window_onsource_strain(on_source_strain, signal_seglen, noise_seglen, window_onsource_flag, window_flag, alpha_window, truncate)
+    
+    # Find the on-source chunk, centered on the trigger time, and the off-source chunk to avoid including the signal in the PSD computation. Since the ACF might be computed on a windowed chunk through an fft, the onsource chunk must be windowed when no truncation is applied and the fft requested, to avoid normalisation inconsistencies.
+    on_source_times, on_source_strain, off_source_strain = compute_on_off_source_strain(times, strain, signal_seglen, index_trigtime)
+    on_source_strain                                     = window_onsource_strain(on_source_strain, signal_seglen, noise_seglen, window_onsource_flag, window_flag, alpha_window, truncate)
 
 
     ###############################
@@ -824,7 +923,7 @@ def load_data(ifo, fname, **kwargs):
     ######################
 
     if not(kwargs['injection-approximant']==''):
-        on_source_strain = add_injection(on_source_strain, strain, starttime, T, ifo, triggertime, index_trigtime, signal_seglen, kwargs)
+        on_source_strain = add_injection(on_source_times, on_source_strain, triggertime, ifo, kwargs)
     else:
         assert not((ifo_datafile=='') and not(download_data_flag)), "No data was passed and no injection has been selected. Exiting."
 
@@ -833,12 +932,12 @@ def load_data(ifo, fname, **kwargs):
     # Finalise output. #
     ####################
 
-    # We are done with the strain.
-    del rawstrain, strain
+    # We are done with the whole strain.
+    del times, rawstrain, strain
 
-    # Return the time series, the covariance matrix and the PSD used in the whitened waveform plot.
+    # Return the: time axis, time series, covariance matrix and PSD used in the whitened waveform plot.
     # Note that the `on_source_strain` is long `signal_seglen` and has not been truncate yet, unlike the ACF, to allow for a variable start time.
-    return on_source_strain, ACF_signal, Covariance_matrix_signal, whitening_PSD
+    return on_source_times, on_source_strain, ACF_signal, Covariance_matrix_signal, whitening_PSD
 
 
 
@@ -849,12 +948,10 @@ def load_data(ifo, fname, **kwargs):
 class detector:
 
     def __init__(self, ifo_name, datafile, **kwargs):
-        self.name                                             = ifo_name
-        self.lal_detector                                     = lal.cached_detector_by_prefix[self.name]
-        self.location                                         = self.lal_detector.location
-        self.time_series, self.acf, self.covariance, self.psd = load_data(self.name, datafile, **kwargs)
-        time_axis_internal                                    = np.linspace(-kwargs['signal-chunksize']/2, kwargs['signal-chunksize']/2, len(self.time_series)+1)[:-1]
-        self.time                                             = kwargs['trigtime'] + time_axis_internal # technically the correct time axis should be centered around the closest datapoint to 'trigtime', i.e. time[index_trigtime], which is slightly different. However in practise this doesn't matter (except for plotting purposes) since in the likelihood 'trigtime' is going to be subtracted again. In other words, the index around which the time_series is centered is the true triggertime used in the likelihood computation.
+        self.name                                                        = ifo_name
+        self.lal_detector                                                = lal.cached_detector_by_prefix[self.name]
+        self.location                                                    = self.lal_detector.location
+        self.time, self.time_series, self.acf, self.covariance, self.psd = load_data(self.name, datafile, **kwargs)
         
         # Save the times and data that will be actually used in the likelihood.
         np.savetxt(kwargs['output']+'/Noise/signal_chunk_times_data_{det}.txt'.format(det=ifo_name), np.column_stack((self.time, self.time_series)))
