@@ -245,7 +245,21 @@ def railing_check(samples, prior_bins, tolerance):
 # From here downwards, NO REVIEW NEEDED. #
 ##########################################
 
-def EsGB_corrections(name):
+def EsGB_corrections(name, dim):
+    
+    # Coefficients of Parspec expansion for gravitational polar-led modes in EdGB, from arXiv:2103.09870, arXiv:2207.11267
+    all_corr = {
+                'domega_220': [-0.03773, -0.1668 , -0.278],
+                'dtau_220'  : [-0.0528 , -0.08   ,  3.914],
+                }
+    if not name in all_corr.keys(): raise ValueError("Currently EdsGB corrections only support the (l,m,n)=(2,2,0) mode deviation.")
+
+    corr = []
+    for i in range(dim+1): corr.append(all_corr[name][i])
+
+    return corr
+
+def EsGB_corrections_Carson_Yagi(name):
     
     if(name=='domega_220'):
         a0_GR =  0.373672
@@ -266,6 +280,7 @@ def EsGB_corrections(name):
         raise ValueError("Tau corrections to be filled yet.")
     else:
         raise ValueError("Currently EdsGB corrections only support the (l,m,n)=(2,2,0) mode deviation.")
+
     return corr
 
 def mtot(m1,m2)        : return m1+m2
